if constructorpieces == nil then
	constructorpieces = {};
	--if there's no global constructorpieces list, create one. the check is to ensure it doesn't get wiped if you order two constructors.
end

function getPixel(x,y,width)
	return (y * width) + x;
	--used for the drawmodule function, waaaaaaaay down.
end

function boxcorners(area,topleft,topright,bottomleft,bottomright,factor)
	local ch1, ch2, ch3, ch4;
	--super simple box area check function. so long as the four arguments are within the area it'll return true, basically.
	if topleft.X >= (area.Pos.X - 12 * factor) and topleft.Y >= (area.Pos.Y - 12 * factor) then
		ch1 = true;
	end
	if topright.X <= (area.Pos.X + 12 * factor) and topright.Y >= (area.Pos.Y - 12 * factor) then
		ch2 = true;
	end
	if bottomleft.X >= (area.Pos.X - 12 * factor) and bottomleft.Y <= (area.Pos.Y + 12 * factor) then
		ch3 = true;
	end
	if bottomright.X <= (area.Pos.X + 12 * factor) and bottomright.Y <= (area.Pos.Y + 12 * factor) then
		ch4 = true;
	end
	return ch1, ch2, ch3, ch4;
end
		

function testcorners(cellnumber,position)
	--slightly fancy function that tells whether the corners of the module to be constructed are within the background modules (defined with mosrotatings of the same size)
	local module = constructorpieces[cellnumber];
	local width,height = module[2].X,module[2].Y;
		
	--local variables that represent each corner of the module
	local tl = Vector(position.X - math.floor(width / 2) % SceneMan.SceneWidth,position.Y - math.floor(height / 2) % SceneMan.SceneHeight);
	local tr = Vector(position.X + math.floor(width / 2) % SceneMan.SceneWidth,position.Y - math.floor(height / 2) % SceneMan.SceneHeight);
	local bl = Vector(position.X - math.floor(width / 2) % SceneMan.SceneWidth,position.Y + math.floor(height / 2) % SceneMan.SceneHeight);
	local br = Vector(position.X + math.floor(width / 2) % SceneMan.SceneWidth,position.Y + math.floor(height / 2) % SceneMan.SceneHeight);

	local check1 = false;
	local check2 = false;
	local check3 = false;
	local check4 = false;

	for particle in MovableMan.Particles do
		if particle.ClassName == "MOSRotating" then
			--don't bother cycling through any particles that aren't mosrotatings
			mosr = ToMOSRotating(particle);
			--tomosrotating is required for isingroup, because MovableMan.Particles only contains basic particle data
			if mosr:IsInGroup("Constructor Backgrounds") == true then
				--if it's inside the group created for the constructor then actually do math:
				for size in string.gmatch(mosr.PresetName,"%d+") do
				--find the size of the mosrotating based on the last letter in the presetname.
					check1,check2,check3,check4 = boxcorners(mosr,tl,tr,bl,br,size);
					break;
				end
			end
		end
		if check1 == true and check2 == true and check3 == true and check4 == true then
			return true;
		end
	end
end

function drawglow(cellnumber,position)
	--simple function to get some info off the global table and draw the module preview glow
	local module = constructorpieces[cellnumber];
	local name = module[1];
	local glow = CreateMOPixel(name,"Constructor.rte");
	local width,height = module[2].X,module[2].Y;
	glow.Pos = position;
	MovableMan:AddParticle(glow);
end

function drawmodule(cellnumber,position,diffy)
	--this is where the magic happens. brilliant function design is mostly TLB and Kyred, I can't claim much.
	local module = constructorpieces[cellnumber];
	local name = module[1];
	local width,height = module[2].X,module[2].Y;
	local deltay,deltax = diffy,0;
	local pixels = module[3];
	--these locals above are set to various attributes from the lua files spit out by bmptoarray. the bracketed numbers refer to cells within the arrays defined in those files.
	local firstrun = true;
		
	if deltay < height then
		for i = (deltay * width), #pixels - 1, 1 do
			deltax = i % width;
			if deltax == 0 and firstrun == false then
				deltay = deltay + 1;
				break;
			end
			--basically this business here just increments over each x pixel while storing the height
			local ppos = Vector(((position.X - math.floor(width / 2)) + deltax) % SceneMan.SceneWidth,((position.Y + math.floor(height / 2)) - (deltay + 1)) % SceneMan.SceneHeight);
			local pixel = pixels[getPixel(deltax + 1,deltay,width)];
			local color = tostring(pixel[1]..","..pixel[2]..","..pixel[3]);
			--color is a vector listed inside the vector that's inside the vector that's inside the vector of modules.
			--you could also call color constructorpieces[cellnumber][3][1]
			if color ~= "255,0,255" then
				--color 255,0,255 means an air pixel, meaning no construction is required. this checks that that's NOT true, allowing it to be built.
				local piece = CreateMOPixel(color,"Constructor.rte");
				piece.Pos = ppos;
				MovableMan:AddParticle(piece);
				piece.ToSettle = true;
				local glow = CreateMOPixel("Pixel Creation Glow","Constructor.rte");
				glow.Pos = ppos;
				MovableMan:AddParticle(glow);
			else
				--if it IS an air pixel, this remove whatever terrain pixel might be there.
				local remover = CreateMOSRotating("Remover","Constructor.rte");
				remover.Pos = ppos;
				MovableMan:AddParticle(remover);
				remover:EraseFromTerrain();
			end
			if i == #pixels - 1 then
				deltay = deltay + 1;
			end
			firstrun = false;
		end
	else
		return false;
	end
	
	return true;
end

function Create(self)
	--set various inits, create timers. modulenumber is the cell number of the current module in the global constructorpieces table.
	self.built = false;
	self.building = false;
	self.firstrun = true;
	self.diffy = 0;
	self.buildtimer = Timer();
	self.movetimer = Timer();
	self.switchtimer = Timer();
	self.modulenumber = 1;
	self.mousepos = self.Pos;
end

function Update(self)
	--draw preview glow if not building
	if self.built == false and self.building == false then
		drawglow(self.modulenumber,self.Pos);
	end
	if self.building == false and self.built == true then
		self.buildtimer:Reset();
		self.diffy = 0;
		self.built = false;
	end

	if self:GetController():IsState(Controller.PRIMARY_ACTION) == true then
		--if you're pressing fire (on any control device)
		if self.buildtimer:IsPastSimMS(1000) == true then
			--make sure you have time to switch modules between constructions
			if self.building == false and self.built == false then
				--if you're not building and you haven't finished building then it tests and starts building
				if testcorners(self.modulenumber,self.Pos) == true then
					self.building = true;
				end
			end
		end
		--if you're building, not done, and have the fire key held, build a module.
		if self.building == true and self.built == false then
			--drawmodule is the drawing function, diffy is the current height of construction
			if drawmodule(self.modulenumber,self.Pos,self.diffy) then
				self.diffy = self.diffy + 1;
			else
				self.built = true;
				self.building = false;
				self.diffy = 0;
			end
		end
	end
	
	---------------control functions past this point
	if self:GetController():IsMouseControlled() == true then
		if self.building == false then
			self.mousepos = self.mousepos + UInputMan:GetMouseMovement(self.Team);
			local newpos = Vector(0,0);
			newpos.X = math.floor((self.mousepos.X / 12)) * 12;
			newpos.Y = math.floor((self.mousepos.Y / 12)) * 12;
			self.Pos.X = newpos.X % SceneMan.SceneWidth;
			self.Pos.Y = newpos.Y % SceneMan.SceneHeight;
		end
		if self.switchtimer:IsPastSimMS(175) == true then
			if self:GetController():IsState(Controller.MOVE_RIGHT) == true then
				self.switchtimer:Reset();
				--constructorpieces is the global table, part of the script files. #tablename returns the number of cells, and since it's an iterated pair we can numerically iterate
				if self.modulenumber < #constructorpieces then
					self.modulenumber = self.modulenumber + 1;
				else
					self.modulenumber = 1;
				end
			elseif self:GetController():IsState(Controller.MOVE_LEFT) == true then
				self.switchtimer:Reset();
				if self.modulenumber > 1 then
					self.modulenumber = self.modulenumber - 1;
				else
					self.modulenumber = #constructorpieces;
				end
			end
		end
	else
		--if we're not mouse controlled (keyboard or gamepad. gamepad gets the short end of the stick; can't switch backwards through module list)
		if self.switchtimer:IsPastSimMS(175) == true then
			if self:GetController():IsState(Controller.PIE_MENU_ACTIVE) == true then
				self.switchtimer:Reset();
				--constructorpieces is the global table, part of the script files. #tablename returns the number of cells, and since it's an iterated pair we can numerically iterate
				if self.modulenumber < #constructorpieces then
					self.modulenumber = self.modulenumber + 1;
				else
					self.modulenumber = 1;
				end
			elseif self:GetController():IsState(Controller.AIM_SHARP) == true then
				self.switchtimer:Reset();
				if self.modulenumber > 1 then
					self.modulenumber = self.modulenumber - 1;
				else
					self.modulenumber = #constructorpieces;
				end
			end
		end
		--basic movement function with some scene width checks. going past scene boundaries is bad news, so there's a few little checks.
		if self.movetimer:IsPastSimMS(40) == true then
			self.movetimer:Reset();
			if self.building == false then
				if self:GetController():IsState(Controller.MOVE_RIGHT) == true then
					if (self.Pos.X + 12) < SceneMan.SceneWidth then
						self.Pos.X = self.Pos.X + 12;
					else
						self.Pos.X = (self.Pos.X + 12 - SceneMan.SceneWidth);
					end
				elseif self:GetController():IsState(Controller.MOVE_LEFT) == true then
					if (self.Pos.X - 12) > 0 then
						self.Pos.X = self.Pos.X - 12;
					else
						self.Pos.X = (self.Pos.X - 12 + SceneMan.SceneWidth);
					end
				end
				if self:GetController():IsState(Controller.MOVE_UP) == true then
					if (self.Pos.Y - 12) > 0 then
						self.Pos.Y = self.Pos.Y - 12;
					else
						self.Pos.Y = (self.Pos.Y - 12 + SceneMan.SceneHeight);
					end
				elseif self:GetController():IsState(Controller.MOVE_DOWN) == true then
					if (self.Pos.Y + 12) < SceneMan.SceneHeight then
						self.Pos.Y = self.Pos.Y + 12;
					else
						self.Pos.Y = (self.Pos.Y + 12 - SceneMan.SceneHeight);
					end
				end
			end
		end
	end
end
